from datetime import datetime

# We use a timestamp format which is compatible with the syntax of CDSTAR bitstream names:
TIMESTAMP_FORMAT = '%Y%m%dT%H%M%SZ'


class RollingBlob(object):
    """
    RollingBlobs are big(gish), versioned files of which only the last couple of versions need to
    be available (such as log files or database dumps).
    """
    def __init__(self, collection=None, name=None, oid=None):
        if not oid:
            if not (collection and name):
                raise ValueError
        else:
            if collection or name:
                raise ValueError
        self.collection = collection
        self.name = name
        self.oid = oid

    @staticmethod
    def parse_timestamp(bsid):
        try:
            return datetime.strptime(bsid.split('_')[-1].split('.')[0], TIMESTAMP_FORMAT)
        except (ValueError, TypeError):
            # Make sure invalid timestamps are sorted as earlier than any valid ones.
            return datetime.strptime('19000101T000000Z', TIMESTAMP_FORMAT)

    def get_object(self, cdstar):
        obj = cdstar.get_object(uid=self.oid)
        if self.oid is None:
            self.oid = obj.id
            obj.metadata = {
                'collection': self.collection,
                'name': self.name,
                'type': self.__class__.__name__,
            }
        else:
            md = obj.metadata.read()
            self.name = md['name']
            self.collection = md['collection']
        return obj

    def add(self, cdstar, fname, suffix='', timestamp=None, mimetype=None):
        if '_' in suffix:
            raise ValueError(suffix)
        timestamp = timestamp or datetime.utcnow()
        if isinstance(timestamp, str):
            timestamp = datetime.strptime(timestamp, TIMESTAMP_FORMAT)
        if suffix and not suffix.startswith('.'):
            suffix = '.' + suffix
        obj = self.get_object(cdstar)
        kw = dict(
            name='{0}_{1}{2}'.format(self.name, timestamp.strftime(TIMESTAMP_FORMAT), suffix),
            fname=fname)
        if mimetype:
            kw['mimetype'] = mimetype
        obj.add_bitstream(**kw)

    def sorted_bitstreams(self, cdstar):
        obj = self.get_object(cdstar)
        return sorted(obj.bitstreams, key=lambda bs: self.parse_timestamp(bs.id), reverse=True)

    def latest(self, cdstar):
        res = self.sorted_bitstreams(cdstar)
        if res:
            return res[0]

    def expunge(self, cdstar, keep=5):
        deleted = 0
        for i, bs in enumerate(self.sorted_bitstreams(cdstar)):
            if i + 1 > keep:
                bs.delete()
                deleted += 1
        return deleted



import time
from collections import OrderedDict, defaultdict
import datetime
import re
import mimetypes
import pathlib

import requests
from requests.packages.urllib3.exceptions import (
    InsecureRequestWarning, InsecurePlatformWarning, SNIMissingWarning,
)
import attr

from pycdstar import media
from pycdstar.catalog import filter_hidden, iter_files
from pycdstar.api import Cdstar
from clldutils.jsonlib import dump, load
from clldutils.misc import format_size

from cdstarcat.resources import RollingBlob

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
requests.packages.urllib3.disable_warnings(InsecurePlatformWarning)
requests.packages.urllib3.disable_warnings(SNIMissingWarning)

mimetypes.add_type('video/mp4', '.mod', strict=False)

OBJID_PATTERN = re.compile('-'.join(['[A-F0-9]{%s}' % n for n in [5, 4, 4, 4, 1]]) + '$')


class WithHumanReadableSize(object):
    @property
    def size_h(self):
        return format_size(getattr(self, 'size', 0))


@attr.s(frozen=True)
class Bitstream(WithHumanReadableSize):
    id = attr.ib()
    size = attr.ib()
    mimetype = attr.ib()
    md5 = attr.ib()
    created = attr.ib()
    modified = attr.ib()

    @property
    def created_datetime(self):
        return datetime.datetime.utcfromtimestamp(self.created / 1e3)

    @property
    def modified_datetime(self):
        return datetime.datetime.utcfromtimestamp(self.modified / 1e3)

    @classmethod
    def fromdict(cls, d):
        return cls(
            d['bitstreamid'],
            d['filesize'],
            d['content-type'],
            d['checksum'],
            d['created'],
            d['last-modified'])

    def asdict(self):
        return OrderedDict([
            ("bitstreamid", self.id),
            ("checksum", self.md5),
            ("created", self.created),
            ("checksum-algorithm", "MD5"),
            ("last-modified", self.modified),
            ("filesize", self.size),
            ("content-type", self.mimetype)
        ])


@attr.s(frozen=True)
class Object(WithHumanReadableSize):
    id = attr.ib()
    bitstreams = attr.ib()
    metadata = attr.ib()

    @property
    def size(self):
        return sum(bs.size for bs in self.bitstreams)

    @property
    def is_special(self):
        return 'warning' in self.metadata

    @classmethod
    def fromdict(cls, id_, d):
        return cls(id_, [Bitstream.fromdict(bs) for bs in d['bitstreams']], d['metadata'])

    def asdict(self):
        return OrderedDict([
            ('bitstreams', [bs.asdict() for bs in self.bitstreams]),
            ('metadata', OrderedDict([(k, v) for k, v in sorted(self.metadata.items())]))]
        )


class Catalog(WithHumanReadableSize):
    """
    A catalog of objects in a CDSTAR instance.

    For operations resulting in changes the Catalog should be used as context manager to
    make sure changes are written to disk.
    """
    def __init__(self, path, cdstar_url=None, cdstar_user=None, cdstar_pwd=None):
        self.path = pathlib.Path(path)
        self.objects = {}
        if self.path.exists():
            self.objects = {i: Object.fromdict(i, d) for i, d in load(self.path).items()}
        self.api = Cdstar(service_url=cdstar_url, user=cdstar_user, password=cdstar_pwd)

    @property
    def md5_to_object(self):
        res = defaultdict(list)
        for obj in self.objects.values():
            for bs in obj.bitstreams:
                res[bs.md5].append(obj)
        return res

    def __enter__(self):
        return self

    def __exit__(self, *args):
        ordered = OrderedDict([(k, v.asdict()) for k, v in sorted(self.objects.items())])
        dump(ordered, self.path, indent=0, separators=(',', ':'))

    def __len__(self):
        """
        :return: The number of objects in the catalog
        """
        return len(self.objects)

    def __iter__(self):
        return iter(self.objects.values())

    def __contains__(self, item):
        """
        Check whether an object or a a bitstream (specified by md5 sum) is in the catalog.

        :param item:
        :return:
        """
        item = getattr(item, 'id', item)
        return (item in self.objects) or (item in self.md5_to_object)

    def __getitem__(self, item):
        item = getattr(item, 'id', item)
        if item in self.objects:
            return self.objects[item]
        md5_to_objects = self.md5_to_object
        if item in md5_to_objects:
            return md5_to_objects[item]
        raise KeyError(item)

    def __setitem__(self, item, obj):
        objid = getattr(item, 'id', item)
        if not OBJID_PATTERN.match(objid):
            raise ValueError('invalid object ID: %s' % objid)
        if not isinstance(obj, Object):
            raise ValueError('invalid object type: %s' % type(obj))
        self.objects[objid] = obj

    @property
    def size(self):
        return sum(obj.size for obj in self)

    def add(self, obj, metadata=None, update=False):
        """
        Add an existing CDSTAR object to the catalog.

        :param obj: A pycdstar.resource.Object instance
        """
        if (obj not in self) or update:
            self[obj.id] = Object.fromdict(
                obj.id,
                dict(
                    metadata=obj.metadata.read() if metadata is None else metadata,
                    bitstreams=[bs._properties for bs in obj.bitstreams]))
            time.sleep(0.1)
            return self.objects[obj.id]

    def add_rollingblob(self, fname, oid=None, collection=None, name=None, **kw):
        rb = RollingBlob(oid=oid, collection=collection, name=name)
        keep = kw.pop('keep', 5)
        rb.add(self.api, fname, **kw)
        rb.expunge(self.api, keep=keep)
        return self.add(rb.get_object(self.api), update=True)

    def remove(self, obj):
        del self.objects[getattr(obj, 'id', obj)]

    def delete(self, obj):
        """
        Delete an object in CDSTAR and remove it from the catalog.

        :param obj: An object ID or an Object instance.
        """
        obj = self.api.get_object(getattr(obj, 'id', obj))
        obj.delete()
        self.remove(obj.id)

    def create(self, path, metadata, filter_=filter_hidden, object_class=None):
        """
        Create objects in CDSTAR and register them in the catalog.

        Note that we guess the mimetype based on the filename extension, using
        `mimetypes.guess_type`. Thus, it is the caller's responsibility to add custom or
        otherwise uncommon types to the list of known types using `mimetypes.add_type`.

        :param path:
        :param metadata:
        :param filter_:
        :return:
        """
        for fname in iter_files(path):
            if not filter_ or filter_(fname):
                created, obj = self._create(fname, metadata, object_class=object_class)
                yield fname, created, obj

    def _create(self, path, metadata, object_class=None):
        mimetype = mimetypes.guess_type(str(path), strict=False)[0] \
            or 'application/octet-stream'
        maintype, subtype = mimetype.split('/')
        cls = object_class or getattr(media, maintype.capitalize(), media.File)
        file_ = cls(path)
        if file_.md5 not in self.md5_to_object:
            obj, md, bitstreams = file_.create_object(self.api, metadata)
            return True, self.add(obj, metadata=md)
        return False, self.md5_to_object[file_.md5][0]

    def update_metadata(self, obj, metadata, mode='merge'):
        objid = getattr(obj, 'id', obj)
        assert OBJID_PATTERN.match(objid) and objid in self
        obj = self.api.get_object(objid)
        md = obj.metadata.read() if mode == 'merge' else {}
        md.update(metadata)
        obj.metadata = md
        return self.add(obj, md, update=True)

    def add_query(self, query, limit=500, offset=0):
        def search(offset):
            time.sleep(0.2)
            return self.api.search(query, index='metadata', limit=limit, offset=offset)

        results = search(offset)
        total_results = 0
        while results:
            for res in results:
                total_results += 1
                self.add(res.resource)
            offset += limit
            results = search(offset)
        return total_results

    def add_objids(self, *objids):
        for objid in objids:
            self.add(self.api.get_object(objid))



"""
Main command line interface of the cdstarcat package.

The basic invocation looks like

    cdstarcat [OPTIONS] <command> [args]

"""
import sys
from collections import Counter
from itertools import groupby, chain
import os

from clldutils.clilib import ArgumentParserWithLogging, command
from clldutils.markup import Table

from cdstarcat import Catalog, OBJID_PATTERN


@command()
def cleanup(args):
    """
    cdstarcat cleanup

    Deletes objects with no bitstreams from CDSTAR and the catalog.
    """
    with _catalog(args) as cat:
        n, d, r = len(cat), [], []
        for obj in cat:
            if not obj.bitstreams:
                if obj.is_special:
                    print('removing {0} from catalog'.format(obj.id))
                    r.append(obj)
                else:
                    print('deleting {0} from CDSTAR'.format(obj.id))
                    d.append(obj)
        for obj in d:
            cat.delete(obj)
        for obj in r:
            cat.remove(obj)
        args.log.info('{0} objects deleted'.format(n - len(cat)))
        return n - len(cat)


@command()
def stats(args):
    """
    cdstarcat stats

    Print summary statistics of bitstreams in the catalog to stdout.
    """
    cat = _catalog(args)
    print('Summary:')
    print('  {0:,} objects with {1:,} bitstreams of total size {2}'.format(
        len(cat), sum(len(obj.bitstreams) for obj in cat), cat.size_h))
    print('  {0} duplicate bitstreams'.format(
        sum(1 for objs in cat.md5_to_object.values() if len(objs) > 1)))
    print('  {0} objects with no bitstreams'.format(
        sum(1 for obj in cat if not obj.bitstreams)))

    print()
    types = Counter(chain(*[[bs.mimetype for bs in obj.bitstreams] for obj in cat]))
    table = Table('maintype', 'subtype', 'bitstreams')
    for maintype, items in groupby(
            sorted(types.items(), key=lambda p: (p[0].split('/')[0], -p[1])),
            lambda p: p[0].split('/')[0]):
        for k, v in items:
            table.append([maintype, k.split('/')[1], v])
    print(table.render(tablefmt='simple'))


@command()
def add(args):
    """
    cdstarcat add SPEC

    Add metadata about objects (specified by SPEC) in CDSTAR to the catalog.
    SPEC: Either a CDSTAR object ID or a query.
    """
    spec = args.args[0]
    with _catalog(args) as cat:
        n = len(cat)
        if OBJID_PATTERN.match(spec):
            cat.add_objids(spec)
        else:
            results = cat.add_query(spec)
            args.log.info('{0} hits for query {1}'.format(results, spec))
        args.log.info('{0} objects added'.format(len(cat) - n))
        return len(cat) - n


@command()
def create(args):
    """
    cdstarcat create PATH

    Create objects in CDSTAR specified by PATH.
    When PATH is a file, a single object (possibly with multiple bitstreams) is created;
    When PATH is a directory, an object will be created for each file in the directory
    (recursing into subdirectories).
    """
    with _catalog(args) as cat:
        for fname, created, obj in cat.create(args.args[0], {}):
            args.log.info('{0} -> {1} object {2.id}'.format(
                fname, 'new' if created else 'existing', obj))


@command()
def delete(args):
    """
    cdstarcat delete OID

    Delete an object specified by OID from CDSTAR.
    """
    with _catalog(args) as cat:
        n = len(cat)
        cat.delete(args.args[0])
        args.log.info('{0} objects deleted'.format(n - len(cat)))
        return n - len(cat)


@command()
def update(args):
    """
    cdstarcat update OID [KEY=VALUE]+

    Update the metadata of an object.
    """
    with _catalog(args) as cat:
        cat.update_metadata(
            args.args[0], dict([arg.split('=', 1) for arg in args.args[1:]]))


def _catalog(args):
    return Catalog(args.catalog, args.url, args.user, args.pwd)


def main():  # pragma: no cover
    parser = ArgumentParserWithLogging(__name__, stats, add, cleanup, create, delete, update)
    for arg in ['catalog', 'url', 'user', 'pwd']:
        envvar = 'CDSTAR_{0}'.format(arg.upper())
        parser.add_argument(
            '--' + arg,
            help="defaults to ${0}".format(envvar),
            default=os.environ.get(envvar))
    sys.exit(parser.main())


if __name__ == "__main__":  # pragma: no cover
    main()



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages


setup(
    name='cdstarcat',
    version="1.0.0",
    description='Manage objects in a CDSTAR instance through a catalog',
    long_description=open("README.md").read(),
    long_description_content_type='text/markdown',
    author='Robert Forkel',
    author_email='forkel@shh.mpg.de',
    url='https://github.com/clld/cdstarcat',
    license="Apache 2",
    zip_safe=False,
    keywords='',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    install_requires=[
        'clldutils>=2.0',
        'pycdstar>=1.0.0',
        'attrs>=18.1',
    ],
    extras_require={
        'dev': [
            'tox',
            'flake8',
            'wheel',
            'twine',
        ],
        'test': [
            'mock',
            'pytest>=3.6',
            'pytest-mock',
            'pytest-cov',
            'coverage>=4.2',
        ],
    },
    entry_points={
        'console_scripts': [
            'cdstarcat=cdstarcat.__main__:main',
        ]
    },
)
